from __future__ import annotations

from collections.abc import Callable, Iterable
from dataclasses import dataclass
import datetime
import logging
from typing import Any

from prettyqt import constants, core, custom_models, gui
from prettyqt.qt import QtCore, QtGui
from prettyqt.utils import treeitem


logger = logging.getLogger(__name__)

SMALL_COL_WIDTH = 120
MEDIUM_COL_WIDTH = 200


@dataclass  # (frozen=True)
class ColumnItem:
    """Determines how an object attribute is shown."""

    name: str
    label: Callable[[treeitem.TreeItem], str] | None = None
    checkstate: Callable[
        [treeitem.TreeItem], constants.CheckStateStr | QtCore.Qt.CheckState | bool
    ] | None = None
    sort_value: Callable[[treeitem.TreeItem], str | float] | None = None
    tooltip: Callable[[treeitem.TreeItem], str] | None = None
    doc: str = "<no help available>"
    col_visible: bool = True
    width: int | str = SMALL_COL_WIDTH
    alignment: Callable | int | constants.AlignmentStr | None = None
    line_wrap: gui.textoption.WordWrapModeStr = "none"
    foreground_color: Callable | str | None = None
    background_color: Callable | str | None = None
    decoration: Callable | QtGui.QIcon | None = None
    font: Callable | QtGui.QFont | str | None = None
    selectable: bool = True
    enabled: bool = True
    editable: bool = False
    checkable: bool = False
    tristate: bool = False
    set_edit: Callable | None = None
    set_checkstate: Callable | None = None
    user_data: dict | Callable | None = None

    def __post_init__(self):
        super().__init__()
        self.model = None

    def get_name(self) -> str:
        return self.name

    def get_flag(self, tree_item):
        flag = constants.NO_FLAGS
        if self.selectable:
            flag |= constants.IS_SELECTABLE
        if self.enabled:
            flag |= constants.IS_ENABLED
        if self.editable:
            flag |= constants.IS_EDITABLE
        if self.checkable:
            flag |= constants.IS_CHECKABLE
        if self.tristate:
            flag |= constants.IS_USER_TRISTATE
        return flag

    def get_label(self, tree_item) -> str:
        match self.label:
            case None:
                return ""
            case Callable():
                return self.label(tree_item)
            case _:
                return self.label

    def get_sort_value(self, tree_item) -> str | int:
        match self.sort_value:
            case None:
                return self.get_label(tree_item)
            case Callable():
                return self.sort_value(tree_item)
            case _:
                return self.sort_value

    def get_user_data(self, tree_item, role):
        match self.user_data:
            case None:
                return ""
            case Callable():
                return self.user_data(tree_item, role)
            case dict():
                return self.user_data[role](tree_item)
            case _:
                raise ValueError(self.user_data)

    def get_tooltip(self, tree_item) -> str:
        match self.tooltip:
            case None:
                return ""
            case Callable():
                return self.tooltip(tree_item)
            case str():
                return self.tooltip
            case _:
                raise ValueError(self.tooltip)

    def get_checkstate(self, tree_item) -> bool | QtCore.Qt.CheckState | None:
        match self.checkstate:
            case None | bool():
                return self.checkstate
            case Callable():
                result = self.checkstate(tree_item)
                if isinstance(result, str):
                    result = constants.CHECK_STATE[result]
                return result
            case _:
                raise ValueError(self.checkstate)

    def set_checkstate_value(
        self,
        tree_item,
        value: bool | QtCore.Qt.CheckState | constants.CheckStateStr | None,
    ):
        match value:
            case str():
                value = constants.CHECK_STATE[value]
            case int():
                value = QtCore.Qt.CheckState(value)
        match self.set_checkstate:
            case None:
                return None
            case Callable():
                self.set_checkstate(tree_item, value)
            case _:
                raise ValueError(self.set_checkstate)

    def set_edit_value(self, tree_item, value: str):
        match self.set_edit:
            case None:
                return None
            case Callable():
                self.set_edit(tree_item, value)
            case _:
                raise ValueError(self.set_edit)

    def get_font(self, tree_item) -> QtGui.QFont | None:
        match self.font:
            case None | QtGui.QFont():
                return self.font
            case Callable():
                return self.font(tree_item)
            case str():
                return QtGui.QFont(self.font)
            case _:
                raise ValueError(self.font)

    def get_foreground(self, tree_item) -> QtGui.QColor | None:
        match self.foreground_color:
            case None | QtGui.QColor() | QtGui.QBrush():
                return self.foreground_color
            case Callable():
                return self.foreground_color(tree_item)
            case _:
                raise ValueError(self.foreground_color)

    def get_background(self, tree_item) -> QtGui.QColor | None:
        match self.background_color:
            case None | QtGui.QColor() | QtGui.QBrush():
                return self.background_color
            case Callable():
                return self.background_color(tree_item)
            case _:
                raise ValueError(self.foreground_color)

    def get_decoration(
        self, tree_item
    ) -> QtGui.QColor | QtGui.QPixmap | QtGui.QIcon | None:
        match self.decoration:
            case None | QtGui.QIcon() | QtGui.QPixmap() | QtGui.QColor():
                return self.decoration
            case Callable():
                return self.decoration(tree_item)
            case _:
                raise ValueError(self.decoration)

    def get_alignment(self, tree_item) -> QtCore.Qt.AlignmentFlag:
        match self.alignment:
            case None:
                return constants.ALIGN_CENTER_LEFT
            case Callable():
                return self.alignment(tree_item)
            case str():
                return constants.ALIGNMENTS[self.alignment]
            case QtCore.Qt.AlignmentFlag():
                return self.alignment
            case _:
                raise ValueError(self.alignment)

    def get_width(self) -> int:
        match self.width:
            case "small":
                return SMALL_COL_WIDTH
            case "medium":
                return MEDIUM_COL_WIDTH
            case int():
                return self.width
            case _:
                raise ValueError(self.width)


class ColumnItemModelMixin:
    def set_columns(self, columns: Iterable[ColumnItem]):
        self._attr_cols = columns
        for col in columns:
            col.model = self

    def data(self, index, role=constants.DISPLAY_ROLE):
        """Return the tree item at the given index and role."""
        if not index.isValid():
            return None

        col = index.column()
        tree_item = self.data_by_index(index)
        col_item = self._attr_cols[col]
        match role:
            case constants.DISPLAY_ROLE | constants.EDIT_ROLE:
                label = col_item.get_label(tree_item)
                if isinstance(label, datetime.datetime):
                    label = str(label)
                return label
            case constants.DECORATION_ROLE:
                return col_item.get_decoration(tree_item)
            case constants.CHECKSTATE_ROLE:
                return col_item.get_checkstate(tree_item)
            case constants.ALIGNMENT_ROLE:
                return col_item.get_alignment(tree_item)
            case constants.FOREGROUND_ROLE:
                return col_item.get_foreground(tree_item)
            case constants.BACKGROUND_ROLE:
                return col_item.get_background(tree_item)
            case constants.FONT_ROLE:
                return col_item.get_font(tree_item)
            case constants.SORT_ROLE:
                return col_item.get_sort_value(tree_item)
            case constants.TOOLTIP_ROLE:
                return col_item.get_tooltip(tree_item)
            case constants.USER_ROLE:
                return tree_item
            case _:
                if int(role) >= int(constants.USER_ROLE):
                    return col_item.get_user_data(tree_item, role)

    def setData(self, index, value, role):
        if not index.isValid():
            return False
        col = index.column()
        tree_item = self.data_by_index(index)
        match role:
            case constants.EDIT_ROLE:
                self._attr_cols[col].set_edit_value(tree_item, value)
                self.dataChanged.emit(index, index)
                return True
            case constants.CHECKSTATE_ROLE:
                self._attr_cols[col].set_checkstate_value(tree_item, value)
                self.dataChanged.emit(index, index)
                return True

    def flags(self, index):
        if not index.isValid():
            return super().flags(index)  # TODO: whats best here?
        col = index.column()
        tree_item = self.data_by_index(index)
        return self._attr_cols[col].get_flag(tree_item)

    def headerData(self, section, orientation, role):
        match orientation, role:
            case constants.HORIZONTAL, constants.DISPLAY_ROLE:
                return self._attr_cols[section].name
            case _, _:
                return None


class ColumnItemModel(ColumnItemModelMixin, custom_models.TreeModel):
    def __init__(
        self,
        obj,
        columns: Iterable[ColumnItem],
        mime_type: str | None = None,
        show_root: bool = True,
        parent: QtCore.QObject | None = None,
        **kwargs,
    ):
        super().__init__(parent, **kwargs)
        self._root_item = treeitem.TreeItem(obj=obj)
        self._show_root = show_root
        self.mime_type = mime_type
        self._attr_cols = []
        self.set_columns(columns)
        self.set_root_item(obj)

    def columnCount(self, parent=None):
        return len(self._attr_cols)


class ColumnTableModel(ColumnItemModelMixin, core.AbstractTableModel):
    def __init__(
        self,
        items: list,
        columns: Iterable[ColumnItem],
        mime_type: str | None = None,
        parent: QtCore.QObject | None = None,
        **kwargs,
    ):
        super().__init__(parent, **kwargs)
        self.items = items
        self.mime_type = mime_type
        self._attr_cols = []
        self.set_columns(columns)

    def rowCount(self, parent=None):
        parent = parent or core.ModelIndex()
        return 0 if parent.isValid() else len(self.items)

    def columnCount(self, parent=None):
        parent = parent or core.ModelIndex()
        return 0 if parent.isValid() else len(self._attr_cols)

    def data_by_index(self, index: core.ModelIndex):
        return self.items[index.row()]

    def setData(self, index, value, role):
        if role == constants.USER_ROLE:
            self.items[index.row()] = value
            self.update_row(index.row())
            return True
        return super().setData(index, value, role)

    def removeRows(self, row: int, count: int, parent):
        end_row = row + count - 1
        with self.remove_rows(row, end_row, parent):
            for i in range(end_row, row - 1, -1):
                self.items.pop(i)
        return True

    def dropMimeData(self, mime_data, action, row, column, parent_index):
        if not mime_data.hasFormat(self.mime_type):
            return False
        # Since we only drop in between items, parent_index must be invalid,
        # and we use the row arg to know where the drop took place.
        if parent_index.isValid():
            return False
        indexes = mime_data.get_json_data(self.mime_type)
        pos = row if row < len(self.items) and row != -1 else len(self.items)
        rem_offset = sum(i <= pos for i in indexes)
        new = [self.items[i] for i in indexes]
        with self.change_layout():
            for i in sorted(indexes, reverse=True):
                self.items.pop(i)
            for item in reversed(new):
                self.items.insert(pos - rem_offset, item)
        return False

    def sort(self, ncol: int, order):
        """Sort table by given column number."""
        is_asc = order == constants.ASCENDING
        if sorter := self._attr_cols[ncol].label:
            with self.change_layout():
                self.items.sort(key=sorter, reverse=is_asc)

    def add(self, item: Any, position: int | None = None):
        """Append provided item to the list."""
        self.add_items(items=[item], position=position)
        return item

    def add_items(self, items: Iterable[Any], position: int | None = None):
        """Append a list of items to the list."""
        if position is None:
            position = len(self.items)
        items = list(items)
        with self.insert_rows(position, position + len(items) - 1):
            for i, _ in enumerate(items):
                self.items.insert(i + position, items[i])
            # self.items.extend(items)
        return items

    def remove_items(self, offsets: Iterable[int]):
        for offset in sorted(offsets, reverse=True):
            self.removeRow(offset)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    test = {}
    colitem = ColumnItem(
        name="Test",
        label=lambda volume: str(volume.get_root_path()),
        checkable=True,
        checkstate=lambda item: getattr(item, "test", False),
        set_checkstate=lambda item, value: setattr(item, "test", value),
        user_data={constants.USER_ROLE + 1: lambda volume: str(volume.get_root_path())},
    )
    items = core.StorageInfo.get_mounted_volumes()
    model = ColumnTableModel(items, [colitem])
    table = widgets.TableView()
    table.setSortingEnabled(True)
    table.set_model(model)
    table.show()
    app.main_loop()



from __future__ import annotations

from collections.abc import Iterator
from typing import Any, Literal

from prettyqt import constants, gui, iconprovider
from prettyqt.qt import QtCore, QtWidgets
from prettyqt.utils import InvalidParamError, bidict, datatypes, get_repr, serializemixin


mod = QtWidgets.QTreeWidgetItem

CHILD_INDICATOR_POLICY = bidict(
    show=mod.ChildIndicatorPolicy.ShowIndicator,
    dont_show=mod.ChildIndicatorPolicy.DontShowIndicator,
    dont_show_when_childless=mod.ChildIndicatorPolicy.DontShowIndicatorWhenChildless,
)

ChildIndicatorPolicyStr = Literal["show", "dont_show", "dont_show_when_childless"]


class TreeWidgetItem(serializemixin.SerializeMixin, QtWidgets.QTreeWidgetItem):
    def __repr__(self):
        return get_repr(self)

    def serialize_fields(self):
        data = [
            self.data(i, constants.USER_ROLE)  # type: ignore
            for i in range(self.columnCount())
        ]
        return dict(
            text=[self.text(i) for i in range(self.columnCount())],
            tool_tip=[self.toolTip(i) for i in range(self.columnCount())],
            status_tip=[self.statusTip(i) for i in range(self.columnCount())],
            checkstate=[self.get_checkstate(i) for i in range(self.columnCount())],
            icon=[self.get_icon(i) for i in range(self.columnCount())],
            data=data,
        )

    def __iter__(self) -> Iterator[QtWidgets.QTreeWidgetItem]:
        return iter(self.child(i) for i in range(self.childCount()))

    def __len__(self):
        return self.childCount()

    def __getitem__(self, index: int) -> QtWidgets.QTreeWidgetItem:
        item = self.child(index)
        if item is None:
            raise KeyError(index)
        return item

    def __delitem__(self, index: int):
        self.takeChild(index)

    def __add__(self, other: QtWidgets.QTreeWidgetItem) -> TreeWidgetItem:
        self.addChild(other)
        return self

    def setChecked(self, column: int, checked: bool):
        self.setCheckState(
            column,
            QtCore.Qt.CheckState.Checked if checked else QtCore.Qt.CheckState.Unchecked,
        )

    def isChecked(self, col: int) -> bool:
        return self.checkState(col) == QtCore.Qt.CheckState.Checked

    def get_children(self, recursive: bool = False) -> list[QtWidgets.QTreeWidgetItem]:
        """Get children of this item.

        recursive option is written iteratively to also support original QTreeWidgetItems.
        """
        if not recursive:
            return [self.child(i) for i in range(self.childCount())]
        results = []
        nodes = [self]
        while nodes:
            items = []
            for node in nodes:
                results.append(node)
                items.extend(node.child(i) for i in range(node.childCount()))
            nodes = items
        return results[1:]

    def get_top_level_items(self) -> list[QtWidgets.QTreeWidgetItem]:
        return [self.topLevelItem(i) for i in range(self.topLevelItemCount())]

    def collapse(self, recursive: bool = False):
        if recursive:
            for i in range(self.childCount()):
                self.child(i).collapse(True)
        self.setExpanded(False)

    def expand(self, recursive: bool = False):
        self.setExpanded(True)
        if recursive:
            for i in range(self.childCount()):
                self.child(i).expand(True)

    def set_size_hint(self, hint: datatypes.SizeType, column: int = 0):
        if isinstance(hint, tuple):
            hint = QtCore.QSize(*hint)
        self.setSizeHint(column, hint)

    def sort_children(self, column: int, descending: bool = False):
        order = constants.DESCENDING if descending else constants.ASCENDING
        self.sortChildren(column, order)

    def set_data(self, data: Any, role: constants.ItemDataRoleStr | int):
        if isinstance(role, str):
            role = constants.ITEM_DATA_ROLE[role]
        super().setData(data, role)

    def get_data(self, role: constants.ItemDataRoleStr | int) -> Any:
        if isinstance(role, str):
            role = constants.ITEM_DATA_ROLE[role]
        return super().data(role)

    def set_icon(self, icon: datatypes.IconType, column: int = 0):
        """Set the icon for the action.

        Args:
            icon: icon to use
            column: column
        """
        icon = iconprovider.get_icon(icon)
        self.setIcon(column, icon)

    def get_background(self, column: int = 0) -> gui.Brush:
        return gui.Brush(self.background(column))

    def get_foreground(self, column: int = 0) -> gui.Brush:
        return gui.Brush(self.foreground(column))

    def get_font(self, column: int = 0) -> gui.Font:
        return gui.Font(self.font(column))

    def get_icon(self, column: int = 0) -> gui.Icon | None:
        icon = self.icon(column)
        return None if icon.isNull() else gui.Icon(icon)

    def set_checkstate(self, state: constants.CheckStateStr, column: int = 0):
        """Set checkstate of the checkbox.

        Args:
            state: checkstate to use
            column: column

        Raises:
            InvalidParamError: invalid checkstate
        """
        if state not in constants.CHECK_STATE:
            raise InvalidParamError(state, constants.CHECK_STATE)
        self.setCheckState(column, constants.CHECK_STATE[state])

    def get_checkstate(self, column: int = 0) -> constants.CheckStateStr:
        """Return checkstate.

        Args:
            column: column

        Returns:
            checkstate
        """
        return constants.CHECK_STATE.inverse[self.checkState(column)]

    def set_child_indicator_policy(self, policy: ChildIndicatorPolicyStr):
        """Set the child indicator policy.

        Args:
            policy: child indicator policy

        Raises:
            InvalidParamError: policy does not exist
        """
        if policy not in CHILD_INDICATOR_POLICY:
            raise InvalidParamError(policy, CHILD_INDICATOR_POLICY)
        self.setChildIndicatorPolicy(CHILD_INDICATOR_POLICY[policy])

    def get_child_indicator_policy(self) -> ChildIndicatorPolicyStr:
        """Return current child indicator policy.

        Returns:
            child indicator policy
        """
        return CHILD_INDICATOR_POLICY.inverse[self.childIndicatorPolicy()]

    def get_text_alignment(self, column: int) -> constants.AlignmentStr:
        return constants.ALIGNMENTS.inverse[self.textAlignment(column)]


if __name__ == "__main__":
    item = TreeWidgetItem()
    print(item.get_text_alignment(0))



"""Module containing custom widget classes."""

from .imageviewer import ImageViewer
from .elidedlabel import ElidedLabel
from .listinput import ListInput
from .booldicttoolbutton import BoolDictToolButton
from .optionalwidget import OptionalWidget
from .collapsibleframe import CollapsibleFrame
from .expandableline import ExpandableLine
from .hierarchicalheaderview import HierarchicalHeaderView
from .keycombinationedit import KeyCombinationEdit
from .singlelinetextedit import SingleLineTextEdit
from .rectedit import RectEdit, RegionEdit
from .enumcombobox import EnumComboBox
from .paletteedit import PaletteEdit
from .cursoredit import CursorEdit
from .iconedit import IconEdit
from .localeedit import LocaleEdit
from .pointedit import PointEdit
from .sizeedit import SizeEdit
from .sizepolicyedit import SizePolicyEdit
from .regexlineedit import RegexLineEdit
from .regexinput import RegexInput
from .mappedcheckbox import MappedCheckBox
from .logtextedit import LogTextEdit
from .enumflagwidget import EnumFlagWidget
from .flagselectionwidget import FlagSelectionWidget
from .stringornumberwidget import StringOrNumberWidget
from .iconlabel import IconLabel
from .iconbrowser import IconBrowser
from .iconwidget import IconWidget
from .flowlayout import FlowLayout
from .borderlayout import BorderLayout
from .completionwidget import CompletionWidget
from .sidebarwidget import SidebarWidget
from .colorcombobox import ColorComboBox
from .colorchooserbutton import ColorChooserButton
from .filechooserbutton import FileChooserButton
from .fontchooserbutton import FontChooserButton
from .inputandslider import InputAndSlider
from .spanslider import SpanSlider
from .labeledslider import LabeledSlider
from .waitingspinner import WaitingSpinner
from .markdownwidget import MarkdownWindow
from .popupinfo import PopupInfo
from .selectionwidget import SelectionWidget
from .codeeditor import CodeEditor
from .roundprogressbar import RoundProgressBar
from .subsequencecompleter import SubsequenceCompleter
from .framelesswindow import FramelessWindow
from .filetree import FileTree

# from .stareditor import StarEditor, StarRating
from .timeline import Timeline, VideoSample
from .standardiconswidget import StandardIconsWidget


__all__ = [
    "ImageViewer",
    "ElidedLabel",
    "ListInput",
    "BoolDictToolButton",
    "OptionalWidget",
    "CollapsibleFrame",
    "CompletionWidget",
    "ExpandableLine",
    "HierarchicalHeaderView",
    "KeyCombinationEdit",
    "SingleLineTextEdit",
    "RegexLineEdit",
    "RegexInput",
    "MappedCheckBox",
    "LogTextEdit",
    "EnumFlagWidget",
    "FlagSelectionWidget",
    "StringOrNumberWidget",
    "IconLabel",
    "IconBrowser",
    "IconWidget",
    "RectEdit",
    "PaletteEdit",
    "CursorEdit",
    "LocaleEdit",
    "IconEdit",
    "PointEdit",
    "SizeEdit",
    "RegionEdit",
    "SizePolicyEdit",
    "FlowLayout",
    "BorderLayout",
    "SidebarWidget",
    "EnumComboBox",
    "ColorComboBox",
    "ColorChooserButton",
    "FileChooserButton",
    "FontChooserButton",
    "InputAndSlider",
    "SpanSlider",
    "LabeledSlider",
    "WaitingSpinner",
    "RoundProgressBar",
    "PopupInfo",
    "SelectionWidget",
    "MarkdownWindow",
    "CodeEditor",
    "Player",
    "Timeline",
    # "StarEditor",
    # "StarRating",
    "VideoSample",
    "RegexEditorWidget",
    "StandardIconsWidget",
    "SubsequenceCompleter",
    "FramelessWindow",
    "FileTree",
]



from __future__ import annotations

import re
import sys

from prettyqt import core


MAX_SIZE = sys.maxsize


class SubsequenceSortFilterProxyModel(core.SortFilterProxyModel):
    """Performs subsequence matching/sorting."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.search_term = ""
        self.filter_patterns = []
        self.filter_patterns_case_sensitive = []
        self.sort_patterns = []

    def set_search_term(self, search_term: str):
        self.search_term = search_term
        self.filter_patterns = []
        self.filter_patterns_case_sensitive = []
        self.sort_patterns = []
        flags = 0 if self.is_filter_case_sensitive() else re.IGNORECASE
        for i in reversed(range(1, len(search_term) + 1)):
            ptrn = f".*{search_term[:i]}.*{search_term[i:]}"
            try:
                self.filter_patterns.append(re.compile(ptrn, flags))
                self.filter_patterns_case_sensitive.append(re.compile(ptrn, 0))
                ptrn = f"{search_term[:i]}.*{search_term[i:]}"
                self.sort_patterns.append(re.compile(ptrn, flags))
            except Exception:
                continue
        self.invalidateRowsFilter()

    def filterAcceptsRow(self, row, index):
        column = self.filterKeyColumn()
        role = self.filterRole()
        idx = self.sourceModel().index(row, column, index)
        completion = self.sourceModel().data(idx, role)
        if (
            completion is None
            or self.search_term is None
            or len(completion) < len(self.search_term)
        ):
            return False
        if len(self.search_term) == 1:
            search_term = self.search_term
            if not self.is_filter_case_sensitive():
                completion = completion.lower()
                search_term = search_term.lower()
            # rank = completion.index(search_term)
            # self.sourceModel().setData(idx, rank, constants.USER_ROLE)
            return search_term in completion
        for pattern, pattern_case, sort_pattern in zip(
            self.filter_patterns,
            self.filter_patterns_case_sensitive,
            self.sort_patterns,
        ):
            if re.match(pattern, completion):
                # compute rank, the lowest rank the closer it is from the
                # completion
                # start = MAX_SIZE
                # for m in sort_pattern.finditer(completion):
                #     start, end = m.span()
                # rank = start + i * 10
                # if re.match(pattern_case, completion):
                #     # favorise completions where case is matched
                #     rank -= 10
                # self.sourceModel().setData(idx, rank, constants.USER_ROLE)
                return True
        return len(self.search_term) == 0


if __name__ == "__main__":
    from prettyqt import widgets
    from prettyqt.custom_models import JsonModel

    app = widgets.app()
    dist = [
        dict(
            a=2,
            b={
                "a": 4,
                "b": [1, 2, 3],
                "jkjkjk": "tekjk",
                "sggg": "tekjk",
                "fdfdf": "tekjk",
                "xxxx": "xxx",
            },
        ),
        6,
        "jkjk",
    ]
    source_model = JsonModel(dist)
    model = SubsequenceSortFilterProxyModel()
    model.setFilterKeyColumn(1)
    model.set_search_term("x")
    model.setSourceModel(source_model)
    table = widgets.TreeView()
    table.setRootIsDecorated(True)
    lineedit = widgets.LineEdit()
    lineedit.value_changed.connect(model.set_search_term)
    widget = widgets.Widget()
    widget.set_layout("vertical")
    widget.box.add(lineedit)
    widget.box.add(table)
    # table.setSortingEnabled(True)
    table.set_model(model)
    widget.show()
    app.main_loop()



from __future__ import annotations

from collections.abc import Callable
import sys
import traceback
from typing import Literal

from prettyqt import gui, iconprovider, widgets
from prettyqt.qt import QtCore, QtWidgets
from prettyqt.utils import InvalidParamError, bidict, datatypes


ICONS = bidict(
    none=QtWidgets.QMessageBox.Icon.NoIcon,
    information=QtWidgets.QMessageBox.Icon.Information,
    warning=QtWidgets.QMessageBox.Icon.Warning,
    critical=QtWidgets.QMessageBox.Icon.Critical,
    question=QtWidgets.QMessageBox.Icon.Question,
)

IconStr = Literal["none", "information", "warning", "critical", "question"]

STANDARD_BUTTON = bidict(
    none=QtWidgets.QMessageBox.StandardButton.NoButton,
    cancel=QtWidgets.QMessageBox.StandardButton.Cancel,
    ok=QtWidgets.QMessageBox.StandardButton.Ok,
    save=QtWidgets.QMessageBox.StandardButton.Save,
    open=QtWidgets.QMessageBox.StandardButton.Open,
    close=QtWidgets.QMessageBox.StandardButton.Close,
    discard=QtWidgets.QMessageBox.StandardButton.Discard,
    apply=QtWidgets.QMessageBox.StandardButton.Apply,
    reset=QtWidgets.QMessageBox.StandardButton.Reset,
    restore_defaults=QtWidgets.QMessageBox.StandardButton.RestoreDefaults,
    help=QtWidgets.QMessageBox.StandardButton.Help,
    save_all=QtWidgets.QMessageBox.StandardButton.SaveAll,
    yes=QtWidgets.QMessageBox.StandardButton.Yes,
    yes_to_all=QtWidgets.QMessageBox.StandardButton.YesToAll,
    no=QtWidgets.QMessageBox.StandardButton.No,
    no_to_all=QtWidgets.QMessageBox.StandardButton.NoToAll,
    abort=QtWidgets.QMessageBox.StandardButton.Abort,
    retry=QtWidgets.QMessageBox.StandardButton.Retry,
    ignore=QtWidgets.QMessageBox.StandardButton.Ignore,
)

StandardButtonStr = Literal[
    "none",
    "cancel",
    "ok",
    "save",
    "open",
    "close",
    "discard",
    "apply",
    "reset",
    "restore_defaults",
    "help",
    "save_all",
    "yes",
    "yes_to_all",
    "no",
    "no_to_all",
    "abort",
    "retry",
    "ignore",
]

BUTTON_ROLE = bidict(
    invalid=QtWidgets.QMessageBox.ButtonRole.InvalidRole,
    accept=QtWidgets.QMessageBox.ButtonRole.AcceptRole,
    reject=QtWidgets.QMessageBox.ButtonRole.RejectRole,
    destructive=QtWidgets.QMessageBox.ButtonRole.DestructiveRole,
    action=QtWidgets.QMessageBox.ButtonRole.ActionRole,
    help=QtWidgets.QMessageBox.ButtonRole.HelpRole,
    yes=QtWidgets.QMessageBox.ButtonRole.YesRole,
    no=QtWidgets.QMessageBox.ButtonRole.NoRole,
    apply=QtWidgets.QMessageBox.ButtonRole.ApplyRole,
    reset=QtWidgets.QMessageBox.ButtonRole.ResetRole,
)

ButtonRoleStr = Literal[
    "invalid",
    "accept",
    "reject",
    "destructive",
    "action",
    "help",
    "yes",
    "no",
    "apply",
    "reset",
]

TEXT_FORMAT = bidict(
    rich=QtCore.Qt.TextFormat.RichText,
    plain=QtCore.Qt.TextFormat.PlainText,
    auto=QtCore.Qt.TextFormat.AutoText,
    markdown=QtCore.Qt.TextFormat.MarkdownText,
)

TextFormatStr = Literal["rich", "plain", "auto", "markdown"]


class MessageBox(widgets.DialogMixin, QtWidgets.QMessageBox):
    def __init__(
        self,
        icon: datatypes.IconType | IconStr = None,
        title: str = "",
        text: str = "",
        informative_text: str = "",
        details: str = "",
        buttons: list[StandardButtonStr] | None = None,
        parent: QtWidgets.QWidget | None = None,
    ):
        super().__init__(parent)
        self.set_icon(icon)
        self.setText(text)
        self.setInformativeText(informative_text)
        self.setWindowTitle(title)
        self.setWindowFlags(
            QtCore.Qt.WindowType.Dialog
            | QtCore.Qt.WindowType.WindowTitleHint
            | QtCore.Qt.WindowType.CustomizeWindowHint
        )
        self.setDetailedText(details)
        if isinstance(buttons, list):
            for b in buttons:
                self.add_button(b)

    @classmethod
    def message(
        cls,
        text: str,
        title: str = "",
        icon: datatypes.IconType = None,
        detail_text: str | None = None,
    ) -> str:
        m = cls("none", title, text)
        m.set_icon(icon)
        if detail_text is not None:
            m.setDetailedText(detail_text)
        return m.show_blocking()

    @classmethod
    def show_exception(cls, exception: Exception):
        exctype, value = sys.exc_info()[:2]
        tb = traceback.format_exc()
        dlg = cls(text=str(value), title=str(exctype), icon="critical", details=tb)
        dlg.show_blocking()

    def set_icon(self, icon: datatypes.IconType | IconStr):
        if icon in ICONS:
            self.setIcon(ICONS[icon])
        else:
            ico = iconprovider.get_icon(icon)
            self.setIconPixmap(ico.get_pixmap(size=64))

    def show_blocking(self) -> StandardButtonStr:
        return STANDARD_BUTTON.inverse[self.main_loop()]

    def get_icon_pixmap(self) -> gui.Pixmap | None:
        pix = self.iconPixmap()
        return None if pix.isNull() else gui.Pixmap(pix)

    def set_standard_buttons(self, buttons: list[StandardButtonStr]):
        flag = self.StandardButton.NoButton
        for val in buttons:
            flag |= STANDARD_BUTTON[val]
        self.setStandardButtons(flag)

    def get_standard_buttons(self) -> list[StandardButtonStr]:
        return STANDARD_BUTTON.get_list(self.standardButtons())

    def add_button(self, button: StandardButtonStr) -> QtWidgets.QPushButton:
        """Add a default button.

        Args:
            button: button to add

        Returns:
            created button

        Raises:
            InvalidParamError: Button type not available
        """
        if button not in STANDARD_BUTTON:
            raise InvalidParamError(button, STANDARD_BUTTON)
        return self.addButton(STANDARD_BUTTON[button])

    def add_custom_button(
        self, button: str, role: ButtonRoleStr, callback: Callable | None = None
    ) -> QtWidgets.QPushButton:
        btn = self.addButton(button, BUTTON_ROLE[role])
        if callback:
            btn.clicked.connect(callback)

    # @classmethod
    # def show_exception(cls, exception):
    #     header = str(exception[0])
    #     error_text = str(exception[1])
    #     widgets.MessageBox.message(error_text, header, "mdi.exclamation")

    def set_text_format(self, text_format: TextFormatStr):
        """Set the text format.

        Args:
            text_format: text format to use

        Raises:
            InvalidParamError: text format does not exist
        """
        if text_format not in TEXT_FORMAT:
            raise InvalidParamError(text_format, TEXT_FORMAT)
        self.setTextFormat(TEXT_FORMAT[text_format])

    def get_text_format(self) -> TextFormatStr:
        """Return current text format.

        Returns:
            text format
        """
        return TEXT_FORMAT.inverse[self.textFormat()]

    def set_escape_button(self, button: StandardButtonStr | QtWidgets.QAbstractButton):
        if isinstance(button, QtWidgets.QAbstractButton):
            self.setEscapeButton(button)
        else:
            self.setEscapeButton(STANDARD_BUTTON[button])

    def set_default_button(self, button: StandardButtonStr | QtWidgets.QPushButton):
        if isinstance(button, QtWidgets.QPushButton):
            self.setDefaultButton(button)
        else:
            self.setDefaultButton(STANDARD_BUTTON[button])


if __name__ == "__main__":
    app = widgets.app()
    ret = MessageBox(icon="warning", title="header", text="text")
    ret.set_standard_buttons(["ok", "cancel"])
    ret.add_button("tt", "accept", callback=lambda: print("click"))
    ret.show()
    print(ret)
    app.main_loop()



from __future__ import annotations

from prettyqt import constants, core, widgets
from prettyqt.qt import QtCore, QtWidgets


class WidgetDelegate(widgets.StyledItemDelegate):
    def __init__(
        self,
        role: QtCore.Qt.ItemDataRole = constants.USER_ROLE,
        parent: QtWidgets.QAbstractItemView | None = None,
    ):
        super().__init__(parent)
        self._widget_role = role
        # self._editors = {}

    def paint(self, painter, option, index):
        value = self._editor_for_index(index)
        if not isinstance(value, QtWidgets.QWidget):
            super().paint(painter, option, index)
            return
        value.setGeometry(option.rect)
        if option.state & widgets.Style.StateFlag.State_MouseOver:
            painter.fillRect(option.rect, option.palette.highlight())
        if option.state & widgets.Style.StateFlag.State_Selected:
            painter.fillRect(option.rect, option.palette.highlight())
        pixmap = value.grab()
        painter.drawPixmap(option.rect, pixmap)
        # super().paint(painter, option, index)

    def _editor_for_index(self, index):
        # using index.data() does not work with PyQt6, it casts widgets to QObjects
        model = index.model()
        editor = model.data(index, self._widget_role)
        return editor
        # key = str((index.row(), index.column()))
        # while (index := index.parent()).isValid():
        #     key += str((index.row(), index.column()))
        # if key not in self._editors:
        #     self._editors[key] = editor
        #     return editor
        # else:
        #     return self._editors[key]

    def createEditor(self, parent, option, index):
        editor = self._editor_for_index(index)
        if isinstance(editor, QtWidgets.QWidget):
            editor = editor.copy()
            editor.setParent(parent)
            editor.setAutoFillBackground(True)
            return editor
        return super().createEditor(parent, option, index)

    def setEditorData(self, editor, index):
        pass

    def setModelData(self, editor, model, index):
        orig = self._editor_for_index(index)
        if isinstance(orig, QtWidgets.QWidget):
            prop = core.MetaObject(orig.metaObject()).get_user_property()
            value = prop.read(editor)
            prop.write(orig, value)

    def sizeHint(self, option, index):
        editor = self._editor_for_index(index)
        if isinstance(editor, QtWidgets.QWidget):
            return editor.sizeHint()
        else:
            return super().sizeHint(option, index)

    def updateEditorGeometry(
        self,
        editor: QtWidgets.QWidget,
        option: QtWidgets.QStyleOptionViewItem,
        index: QtCore.QModelIndex,
    ):
        pass


if __name__ == "__main__":
    app = widgets.app()
    table_widget = widgets.TableWidget(2, 2)
    # table_widget.set_delegate(StarDelegate(), column=1)
    table_widget.setEditTriggers(table_widget.EditTrigger.AllEditTriggers)
    table_widget.set_selection_behavior("rows")
    table_widget.setHorizontalHeaderLabels(["Title", "Rating"])
    types = dict(
        color=widgets.PlainTextEdit(),
        time=widgets.DateEdit(),
    )
    for i, (k, v) in enumerate(types.items()):
        item_1 = widgets.TableWidgetItem(k)
        item_2 = widgets.TableWidgetItem()
        item_2.setData(constants.DISPLAY_ROLE, v)
        item_2.setData(constants.USER_ROLE, v)
        table_widget[i, 0] = item_1
        table_widget[i, 1] = item_2

    delegate = WidgetDelegate(parent=table_widget)
    table_widget.set_delegate(delegate, column=1)

    table_widget.resizeColumnsToContents()
    table_widget.resize(500, 300)
    table_widget.show()
    with app.debug_mode():
        app.main_loop()



from __future__ import annotations

import contextlib
from typing import Literal

from prettyqt import constants, core
from prettyqt.qt import QtCore, QtWidgets
from prettyqt.utils import bidict


CHECK_INDEX_OPTIONS = bidict(
    none=QtCore.QAbstractItemModel.CheckIndexOption.NoOption,
    index_is_valid=QtCore.QAbstractItemModel.CheckIndexOption.IndexIsValid,
    do_not_use_parent=QtCore.QAbstractItemModel.CheckIndexOption.DoNotUseParent,
    parent_is_invalid=QtCore.QAbstractItemModel.CheckIndexOption.ParentIsInvalid,
)

LAYOUT_CHANGE_HINT = bidict(
    none=QtCore.QAbstractItemModel.LayoutChangeHint.NoLayoutChangeHint,
    vertical_sort=QtCore.QAbstractItemModel.LayoutChangeHint.VerticalSortHint,
    horizontal_sort=QtCore.QAbstractItemModel.LayoutChangeHint.HorizontalSortHint,
)


class AbstractItemModelMixin(core.ObjectMixin):
    DEFAULT_FLAGS = (
        constants.DRAG_ENABLED | constants.IS_ENABLED | constants.IS_SELECTABLE
    )

    def __repr__(self):
        return f"{type(self).__name__}: {self.rowCount()} rows"

    def __len__(self) -> int:
        """Return amount of rows."""
        return self.rowCount()

    def __add__(
        self, other: QtCore.QAbstractItemModel
    ) -> core.ConcatenateTablesProxyModel:
        proxy = core.ConcatenateTablesProxyModel()
        proxy.addSourceModel(self)
        proxy.addSourceModel(other)
        return proxy

    @classmethod
    def ci(
        cls,
        index_is_valid: bool = False,
        do_not_use_parent: bool = False,
        parent_is_invalid: bool = False,
    ):
        def inner(method):
            def wrapper(
                ref: AbstractItemModelMixin, index: QtCore.QModelIndex, *args, **kwargs
            ):
                if ref.check_index(
                    index, index_is_valid, do_not_use_parent, parent_is_invalid
                ):
                    return method(ref, index, *args, **kwargs)
                else:
                    raise TypeError("Invalid index")

            return wrapper

        return inner

    def check_index(
        self,
        index: QtCore.QModelIndex,
        index_is_valid: bool = False,
        do_not_use_parent: bool = False,
        parent_is_invalid: bool = False,
    ) -> bool:
        flag = QtCore.QAbstractItemModel.CheckIndexOption.NoOption
        if index_is_valid:
            flag |= CHECK_INDEX_OPTIONS["index_is_valid"]  # type: ignore
        if do_not_use_parent:
            flag |= CHECK_INDEX_OPTIONS["do_not_use_parent"]  # type: ignore
        if parent_is_invalid:
            flag |= CHECK_INDEX_OPTIONS["parent_is_invalid"]  # type: ignore
        check_flag = QtCore.QAbstractItemModel.CheckIndexOption(0) | flag  # type: ignore
        return self.checkIndex(index, check_flag)  # type: ignore

    @contextlib.contextmanager
    def change_layout(self):
        """Context manager to change the layout.

        wraps calls with correct signals
        emitted at beginning: layoutAboutToBeChanged
        emitted at end: layoutChanged

        """
        self.layoutAboutToBeChanged.emit()
        yield None
        self.layoutChanged.emit()

    @contextlib.contextmanager
    def reset_model(self):
        """Context manager to reset the model.

        wraps calls with correct signals
        emitted at beginning: beginResetModel
        emitted at end: endResetModel

        """
        self.beginResetModel()
        yield None
        self.endResetModel()

    def update_row(self, row: int):
        start_index = self.index(row, 0)
        end_index = self.index(row, self.columnCount() - 1)
        self.dataChanged.emit(start_index, end_index)

    def update_all(self):
        top_left = self.index(0, 0)
        bottom_right = self.index(self.rowCount() - 1, self.columnCount() - 1)
        self.dataChanged.emit(top_left, bottom_right)

    @contextlib.contextmanager
    def remove_row(self, row: int, parent: QtCore.QModelIndex | None = None):
        parent = QtCore.QModelIndex() if parent is None else parent
        self.beginRemoveRows(parent, row, row)
        yield None
        self.endRemoveRows()

    @contextlib.contextmanager
    def remove_rows(
        self,
        first: int | None = None,
        last: int | None = None,
        parent: QtCore.QModelIndex | None = None,
    ):
        parent = QtCore.QModelIndex() if parent is None else parent
        first = first if first is not None else 0
        last = last if last is not None else self.rowCount()
        self.beginRemoveRows(parent, first, last)
        yield None
        self.endRemoveRows()

    @contextlib.contextmanager
    def remove_columns(
        self,
        first: int | None = None,
        last: int | None = None,
        parent: QtCore.QModelIndex | None = None,
    ):
        parent = QtCore.QModelIndex() if parent is None else parent
        first = first if first is not None else 0
        last = last if last is not None else self.rowCount()
        self.beginRemoveColumns(parent, first, last)
        yield None
        self.endRemoveColumns()

    @contextlib.contextmanager
    def insert_row(self, row: int, parent: QtCore.QModelIndex | None = None):
        parent = QtCore.QModelIndex() if parent is None else parent
        self.beginInsertRows(parent, row, row)
        yield None
        self.endInsertRows()

    @contextlib.contextmanager
    def insert_rows(
        self,
        first: int | None = None,
        last: int | None = None,
        parent: QtCore.QModelIndex | None = None,
    ):
        parent = QtCore.QModelIndex() if parent is None else parent
        first = first if first is not None else 0
        last = last if last is not None else self.rowCount()
        self.beginInsertRows(parent, first, last)
        yield None
        self.endInsertRows()

    @contextlib.contextmanager
    def append_rows(self, num_rows: int, parent: QtCore.QModelIndex | None = None):
        parent = QtCore.QModelIndex() if parent is None else parent
        row_count = self.rowCount()
        self.beginInsertRows(parent, row_count, row_count + num_rows - 1)
        yield None
        self.endInsertRows()

    @contextlib.contextmanager
    def insert_columns(
        self,
        first: int | None = None,
        last: int | None = None,
        parent: QtCore.QModelIndex | None = None,
    ):
        parent = QtCore.QModelIndex() if parent is None else parent
        first = first if first is not None else 0
        last = last if last is not None else self.rowCount()
        self.beginInsertColumns(parent, first, last)
        yield None
        self.endInsertColumns()

    def force_reset(self):
        self.beginResetModel()
        self.endResetModel()

    def force_layoutchange(self):
        self.layoutAboutToBeChanged.emit()
        self.layoutChanged.emit()

    def get_role_names(self) -> dict[int, str]:
        return {i: v.data().decode() for i, v in self.roleNames().items()}

    def transpose(
        self, parent: QtWidgets.QWidget | None = None
    ) -> core.TransposeProxyModel:
        # PySide6 needs widget parent here
        parent = parent or self.parent()
        if parent is None:
            raise ValueError("needs parent!")
        proxy = core.TransposeProxyModel(parent=parent)
        proxy.setSourceModel(self)
        return proxy

    def get_search_proxy(
        self, search_type: Literal["regular", "fuzzy"] = "regular", **kwargs
    ):
        match search_type:
            case "regular":
                proxy = core.SortFilterProxyModel(**kwargs)
            case "fuzzy":
                from prettyqt import custom_models

                proxy = custom_models.FuzzyFilterProxyModel(**kwargs)
            case _:
                raise ValueError(search_type)
        proxy.setSourceModel(self)
        return proxy


class AbstractItemModel(AbstractItemModelMixin, QtCore.QAbstractItemModel):
    pass



from __future__ import annotations

import logging
import traceback

from prettyqt import core, custom_models, widgets
from prettyqt.utils import signallogger


logger = logging.getLogger(__name__)


COL_LEVEL_NAME = custom_models.ColumnItem(
    name="Level.",
    doc="Level.",
    label=lambda x: x.levelname,
)
COL_FILENAME = custom_models.ColumnItem(
    name="Filename",
    doc="Filename.",
    label=lambda x: x.filename,
)
COL_FUNCTION_NAME = custom_models.ColumnItem(
    name="Function name",
    doc="Function name.",
    label=lambda x: x.funcName,
)
COL_MODULE = custom_models.ColumnItem(
    name="Module",
    doc="Module",
    label=lambda x: x.module,
)
COL_CREATED = custom_models.ColumnItem(
    name="Created",
    doc="Created.",
    label=lambda x: str(core.DateTime.from_seconds(x.created)),
)
COL_LINE_NO = custom_models.ColumnItem(
    name="Line no",
    doc="Line number.",
    label=lambda x: str(x.lineno),
)
# COL_MSECS = custom_models.ColumnItem(
#     name="Msecs",
#     doc="Millisecons",
#     label=lambda x: str(x.msecs),
# )
COL_PROCESS = custom_models.ColumnItem(
    name="Process",
    doc="Process",
    label=lambda x: str(x.process),
)
COL_THREAD = custom_models.ColumnItem(
    name="Thread",
    doc="Thread.",
    label=lambda x: str(x.thread),
)
COL_THREAD_NAME = custom_models.ColumnItem(
    name="Thread name",
    doc="Thread name.",
    label=lambda x: x.threadName or "",
)
COL_PROCESS_NAME = custom_models.ColumnItem(
    name="Process name",
    doc="Process name.",
    label=lambda x: x.processName or "",
)
COL_RELATIVE_CREATED = custom_models.ColumnItem(
    name="Relative created",
    doc="Relative created.",
    label=lambda x: str(x.relativeCreated),
)
COL_NAME = custom_models.ColumnItem(
    name="Name",
    doc="Name",
    label=lambda x: x.name,
)
COL_PATH_NAME = custom_models.ColumnItem(
    name="Path Name",
    doc="Path name",
    label=lambda x: x.pathname,
)

COL_MESSAGE = custom_models.ColumnItem(
    name="Message",
    doc="Message",
    label=lambda x: traceback.format_exc()
    if isinstance(x.msg, Exception)
    else x.msg % x.args,
)


COLUMNS = [
    COL_CREATED,
    COL_LEVEL_NAME,
    COL_MESSAGE,
    COL_FILENAME,
    COL_FUNCTION_NAME,
    COL_MODULE,
    COL_LINE_NO,
    # COL_MSECS,
    COL_PROCESS,
    COL_THREAD,
    COL_THREAD_NAME,
    COL_PROCESS_NAME,
    # COL_RELATIVE_CREATED,
    COL_NAME,
    COL_PATH_NAME,
]


class LogRecordModel(custom_models.ColumnTableModel):
    def __init__(self, logger, level=logging.DEBUG, *args, **kwargs):
        super().__init__(items=[], columns=COLUMNS)
        self.handler = signallogger.SignalLogger()
        self.handler.signals.log_record.connect(self.add)
        core.CoreApplication.call_on_exit(lambda: logger.removeHandler(self.handler))
        self.handler.setLevel(level)
        logger.addHandler(self.handler)


if __name__ == "__main__":
    logger = logging.getLogger()

    app = widgets.app()
    view = widgets.TreeView()
    model = LogRecordModel(logger, parent=view)

    w = widgets.Widget()
    w.set_layout("vertical")
    widget = widgets.TableView()
    widget.set_model(model)

    def raise_exc():
        try:
            raise Exception("test")
        except Exception as e:
            logger.exception(e)

    w.box.add(widgets.PushButton("Raise", clicked=raise_exc))
    w.box.add(widgets.PushButton("Debug", clicked=lambda: logger.debug("Debug")))
    w.box.add(widgets.PushButton("Info", clicked=lambda: logger.info("Info")))
    w.box.add(widgets.PushButton("Warning", clicked=lambda: logger.warning("Warning")))
    w.box.add(widgets.PushButton("Critical", clicked=lambda: logger.critical("Critical")))
    w.box.add(widget)
    w.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import core, custom_models


loc = core.Locale()

COL_ROOT_PATH = custom_models.ColumnItem(
    name="Root path",
    doc="Root path",
    label=lambda volume: str(volume.get_root_path()),
)

COL_VOLUME_NAME = custom_models.ColumnItem(
    name="Volume name",
    doc="Volume name",
    label=lambda volume: volume.name(),
)

COL_DEVICE = custom_models.ColumnItem(
    name="Device",
    doc="Device",
    label=lambda volume: volume.get_device(),
)

COL_FILE_SYSTEM = custom_models.ColumnItem(
    name="File system",
    doc="File system",
    label=lambda volume: volume.get_file_system_type(),
)

COL_TOTAL = custom_models.ColumnItem(
    name="Total",
    doc="Total",
    label=lambda volume: loc.get_formatted_data_size(volume.bytesTotal()),
)

COL_FREE = custom_models.ColumnItem(
    name="Free",
    doc="Free",
    label=lambda volume: loc.get_formatted_data_size(volume.bytesFree()),
)

COL_AVAILABLE = custom_models.ColumnItem(
    name="Available",
    doc="Available",
    label=lambda volume: loc.get_formatted_data_size(volume.bytesAvailable()),
)

COL_READY = custom_models.ColumnItem(
    name="Available",
    doc="Available",
    label=None,
    checkstate=lambda volume: volume.isReady(),
)

COL_READ_ONLY = custom_models.ColumnItem(
    name="Read-only",
    doc="Read-only",
    label=None,
    checkstate=lambda volume: volume.isReadOnly(),
)

COL_VALID = custom_models.ColumnItem(
    name="Valid",
    doc="Valid",
    label=None,
    checkstate=lambda volume: volume.isValid(),
)

COLUMNS = [
    COL_ROOT_PATH,
    COL_VOLUME_NAME,
    COL_DEVICE,
    COL_FILE_SYSTEM,
    COL_TOTAL,
    COL_FREE,
    COL_AVAILABLE,
    COL_READY,
    COL_READ_ONLY,
    COL_VALID,
]


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    view = widgets.TreeView()
    volumes = core.StorageInfo.get_mounted_volumes()
    model = custom_models.ColumnTableModel(volumes, COLUMNS, parent=view)
    view.setModel(model)
    view.resize(640, 480)
    view.set_selection_behavior("rows")
    view.adapt_sizes()
    view.show()
    app.main_loop()
